package blobovniczatree

import (
	"context"
	"sync"
	"testing"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/object"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/internal/blobstortest"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/shard/mode"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/logger/test"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	"github.com/stretchr/testify/require"
	"golang.org/x/sync/errgroup"
)

func TestBlobovniczaTreeSchemaRebuild(t *testing.T) {
	t.Parallel()

	t.Run("width increased", func(t *testing.T) {
		t.Parallel()
		testBlobovniczaTreeRebuildHelper(t, 2, 2, 2, 3, false)
	})

	t.Run("width reduced", func(t *testing.T) {
		t.Parallel()
		testBlobovniczaTreeRebuildHelper(t, 2, 2, 2, 1, true)
	})

	t.Run("depth increased", func(t *testing.T) {
		t.Parallel()
		testBlobovniczaTreeRebuildHelper(t, 1, 2, 2, 2, true)
	})

	t.Run("depth reduced", func(t *testing.T) {
		t.Parallel()
		testBlobovniczaTreeRebuildHelper(t, 2, 2, 1, 2, true)
	})
}

func TestBlobovniczaTreeFillPercentRebuild(t *testing.T) {
	t.Parallel()

	t.Run("no rebuild by fill percent", func(t *testing.T) {
		t.Parallel()

		dir := t.TempDir()
		b := NewBlobovniczaTree(
			context.Background(),
			WithLogger(test.NewLogger(t)),
			WithObjectSizeLimit(64*1024),
			WithBlobovniczaShallowWidth(1), // single directory
			WithBlobovniczaShallowDepth(1),
			WithRootPath(dir),
			WithBlobovniczaSize(100*1024), // 100 KB limit for each blobovnicza
			WithWaitBeforeDropDB(0),
			WithOpenedCacheSize(1000),
			WithMoveBatchSize(3))
		require.NoError(t, b.Open(mode.ComponentReadWrite))
		require.NoError(t, b.Init())

		storageIDs := make(map[oid.Address][]byte)
		for range 100 {
			obj := blobstortest.NewObject(64 * 1024) // 64KB object
			data, err := obj.Marshal()
			require.NoError(t, err)
			var prm common.PutPrm
			prm.Address = object.AddressOf(obj)
			prm.RawData = data
			res, err := b.Put(context.Background(), prm)
			require.NoError(t, err)
			storageIDs[prm.Address] = res.StorageID
		}
		metaStub := &storageIDUpdateStub{
			storageIDs: storageIDs,
			guard:      &sync.Mutex{},
		}
		rRes, err := b.Rebuild(context.Background(), common.RebuildPrm{
			MetaStorage:   metaStub,
			WorkerLimiter: &rebuildLimiterStub{},
			FillPercent:   60,
		})
		require.NoError(t, err)
		dataMigrated := rRes.ObjectsMoved > 0 || rRes.FilesRemoved > 0 || metaStub.updatedCount > 0
		require.False(t, dataMigrated)

		for addr, storageID := range storageIDs {
			var gPrm common.GetPrm
			gPrm.Address = addr
			gPrm.StorageID = storageID
			_, err := b.Get(context.Background(), gPrm)
			require.NoError(t, err)
		}

		require.NoError(t, b.Close(context.Background()))
	})

	t.Run("no rebuild single db", func(t *testing.T) {
		t.Parallel()

		dir := t.TempDir()
		b := NewBlobovniczaTree(
			context.Background(),
			WithLogger(test.NewLogger(t)),
			WithObjectSizeLimit(64*1024),
			WithBlobovniczaShallowWidth(1), // single directory
			WithBlobovniczaShallowDepth(1),
			WithRootPath(dir),
			WithBlobovniczaSize(100*1024), // 100 KB soft limit for each blobovnicza
			WithWaitBeforeDropDB(0),
			WithOpenedCacheSize(1000),
			WithMoveBatchSize(3))
		require.NoError(t, b.Open(mode.ComponentReadWrite))
		require.NoError(t, b.Init())

		storageIDs := make(map[oid.Address][]byte)
		obj := blobstortest.NewObject(64 * 1024) // 64KB object
		data, err := obj.Marshal()
		require.NoError(t, err)
		var prm common.PutPrm
		prm.Address = object.AddressOf(obj)
		prm.RawData = data
		res, err := b.Put(context.Background(), prm)
		require.NoError(t, err)
		storageIDs[prm.Address] = res.StorageID
		metaStub := &storageIDUpdateStub{
			storageIDs: storageIDs,
			guard:      &sync.Mutex{},
		}
		rRes, err := b.Rebuild(context.Background(), common.RebuildPrm{
			MetaStorage:   metaStub,
			WorkerLimiter: &rebuildLimiterStub{},
			FillPercent:   90, // 64KB / 100KB = 64%
		})
		require.NoError(t, err)
		dataMigrated := rRes.ObjectsMoved > 0 || rRes.FilesRemoved > 0 || metaStub.updatedCount > 0
		require.False(t, dataMigrated)

		for addr, storageID := range storageIDs {
			var gPrm common.GetPrm
			gPrm.Address = addr
			gPrm.StorageID = storageID
			_, err := b.Get(context.Background(), gPrm)
			require.NoError(t, err)
		}

		require.NoError(t, b.Close(context.Background()))
	})

	t.Run("rebuild by fill percent", func(t *testing.T) {
		t.Parallel()

		dir := t.TempDir()
		b := NewBlobovniczaTree(
			context.Background(),
			WithLogger(test.NewLogger(t)),
			WithObjectSizeLimit(64*1024),
			WithBlobovniczaShallowWidth(1), // single directory
			WithBlobovniczaShallowDepth(1),
			WithRootPath(dir),
			WithBlobovniczaSize(100*1024), // 100 KB limit for each blobovnicza
			WithWaitBeforeDropDB(0),
			WithOpenedCacheSize(1000),
			WithMoveBatchSize(3))
		require.NoError(t, b.Open(mode.ComponentReadWrite))
		require.NoError(t, b.Init())

		storageIDs := make(map[oid.Address][]byte)
		toDelete := make(map[oid.Address][]byte)
		for i := range 100 { // 2 objects for one blobovnicza, so 50 DBs total will be created
			obj := blobstortest.NewObject(64 * 1024)
			data, err := obj.Marshal()
			require.NoError(t, err)
			var prm common.PutPrm
			prm.Address = object.AddressOf(obj)
			prm.RawData = data
			res, err := b.Put(context.Background(), prm)
			require.NoError(t, err)
			storageIDs[prm.Address] = res.StorageID
			if i%2 == 1 {
				toDelete[prm.Address] = res.StorageID
			}
		}
		for addr, storageID := range toDelete {
			var prm common.DeletePrm
			prm.Address = addr
			prm.StorageID = storageID
			_, err := b.Delete(context.Background(), prm)
			require.NoError(t, err)
		}
		metaStub := &storageIDUpdateStub{
			storageIDs: storageIDs,
			guard:      &sync.Mutex{},
		}
		rRes, err := b.Rebuild(context.Background(), common.RebuildPrm{
			MetaStorage:   metaStub,
			WorkerLimiter: &rebuildLimiterStub{},
			FillPercent:   80,
		})
		require.NoError(t, err)
		require.Equal(t, uint64(49), rRes.FilesRemoved)
		require.Equal(t, uint64(49), rRes.ObjectsMoved) // 49 DBs with 1 objects
		require.Equal(t, uint64(49), metaStub.updatedCount)

		for addr, storageID := range storageIDs {
			if _, found := toDelete[addr]; found {
				continue
			}
			var gPrm common.GetPrm
			gPrm.Address = addr
			gPrm.StorageID = storageID
			_, err := b.Get(context.Background(), gPrm)
			require.NoError(t, err)
		}

		require.NoError(t, b.Close(context.Background()))
	})

	t.Run("rebuild by overflow", func(t *testing.T) {
		t.Parallel()

		dir := t.TempDir()
		b := NewBlobovniczaTree(
			context.Background(),
			WithLogger(test.NewLogger(t)),
			WithObjectSizeLimit(64*1024),
			WithBlobovniczaShallowWidth(1), // single directory
			WithBlobovniczaShallowDepth(1),
			WithRootPath(dir),
			WithBlobovniczaSize(100*1024), // 100 KB limit for each blobovnicza
			WithWaitBeforeDropDB(0),
			WithOpenedCacheSize(1000),
			WithMoveBatchSize(3))
		require.NoError(t, b.Open(mode.ComponentReadWrite))
		require.NoError(t, b.Init())

		storageIDs := make(map[oid.Address][]byte)
		for range 100 { // 2 objects for one blobovnicza, so 50 DBs total will be created
			obj := blobstortest.NewObject(64 * 1024)
			data, err := obj.Marshal()
			require.NoError(t, err)
			var prm common.PutPrm
			prm.Address = object.AddressOf(obj)
			prm.RawData = data
			res, err := b.Put(context.Background(), prm)
			require.NoError(t, err)
			storageIDs[prm.Address] = res.StorageID
		}
		metaStub := &storageIDUpdateStub{
			storageIDs: storageIDs,
			guard:      &sync.Mutex{},
		}
		require.NoError(t, b.Close(context.Background()))
		b = NewBlobovniczaTree(
			context.Background(),
			WithLogger(test.NewLogger(t)),
			WithObjectSizeLimit(64*1024),
			WithBlobovniczaShallowWidth(1),
			WithBlobovniczaShallowDepth(1),
			WithRootPath(dir),
			WithBlobovniczaSize(50*1024), // 50 KB limit for each blobovnicza
			WithWaitBeforeDropDB(0),
			WithOpenedCacheSize(1000),
			WithMoveBatchSize(3))
		require.NoError(t, b.Open(mode.ComponentReadWrite))
		require.NoError(t, b.Init())

		rRes, err := b.Rebuild(context.Background(), common.RebuildPrm{
			MetaStorage:   metaStub,
			WorkerLimiter: &rebuildLimiterStub{},
			FillPercent:   80,
		})
		require.NoError(t, err)
		require.Equal(t, uint64(49), rRes.FilesRemoved)
		require.Equal(t, uint64(98), rRes.ObjectsMoved) // 49 DBs with 2 objects
		require.Equal(t, uint64(98), metaStub.updatedCount)

		for addr, storageID := range storageIDs {
			var gPrm common.GetPrm
			gPrm.Address = addr
			gPrm.StorageID = storageID
			_, err := b.Get(context.Background(), gPrm)
			require.NoError(t, err)
		}

		require.NoError(t, b.Close(context.Background()))
	})
}

func TestBlobovniczaTreeRebuildLargeObject(t *testing.T) {
	t.Parallel()

	dir := t.TempDir()
	b := NewBlobovniczaTree(
		context.Background(),
		WithLogger(test.NewLogger(t)),
		WithObjectSizeLimit(64*1024), // 64KB object size limit
		WithBlobovniczaShallowWidth(5),
		WithBlobovniczaShallowDepth(2), // depth = 2
		WithRootPath(dir),
		WithBlobovniczaSize(100*1024),
		WithWaitBeforeDropDB(0),
		WithOpenedCacheSize(1000),
		WithMoveBatchSize(3))
	require.NoError(t, b.Open(mode.ComponentReadWrite))
	require.NoError(t, b.Init())

	obj := blobstortest.NewObject(64 * 1024) // 64KB object
	data, err := obj.Marshal()
	require.NoError(t, err)
	var prm common.PutPrm
	prm.Address = object.AddressOf(obj)
	prm.RawData = data
	res, err := b.Put(context.Background(), prm)
	require.NoError(t, err)

	storageIDs := make(map[oid.Address][]byte)
	storageIDs[prm.Address] = res.StorageID

	require.NoError(t, b.Close(context.Background()))

	b = NewBlobovniczaTree(
		context.Background(),
		WithLogger(test.NewLogger(t)),
		WithObjectSizeLimit(32*1024), // 32KB object size limit
		WithBlobovniczaShallowWidth(5),
		WithBlobovniczaShallowDepth(3), // depth = 3
		WithRootPath(dir),
		WithBlobovniczaSize(100*1024),
		WithWaitBeforeDropDB(0),
		WithOpenedCacheSize(1000),
		WithMoveBatchSize(3))
	require.NoError(t, b.Open(mode.ComponentReadWrite))
	require.NoError(t, b.Init())

	metaStub := &storageIDUpdateStub{
		storageIDs: storageIDs,
		guard:      &sync.Mutex{},
	}
	var rPrm common.RebuildPrm
	rPrm.MetaStorage = metaStub
	rPrm.WorkerLimiter = &rebuildLimiterStub{}
	rPrm.FillPercent = 1
	rRes, err := b.Rebuild(context.Background(), rPrm)
	require.NoError(t, err)
	dataMigrated := rRes.ObjectsMoved > 0 || rRes.FilesRemoved > 0 || metaStub.updatedCount > 0
	require.True(t, dataMigrated)

	for addr, storageID := range storageIDs {
		var gPrm common.GetPrm
		gPrm.Address = addr
		gPrm.StorageID = storageID
		_, err := b.Get(context.Background(), gPrm)
		require.NoError(t, err)
	}

	require.NoError(t, b.Close(context.Background()))
}

func testBlobovniczaTreeRebuildHelper(t *testing.T, sourceDepth, sourceWidth, targetDepth, targetWidth uint64, shouldMigrate bool) {
	dir := t.TempDir()
	b := NewBlobovniczaTree(
		context.Background(),
		WithLogger(test.NewLogger(t)),
		WithObjectSizeLimit(2048),
		WithBlobovniczaShallowWidth(sourceWidth),
		WithBlobovniczaShallowDepth(sourceDepth),
		WithRootPath(dir),
		WithBlobovniczaSize(100*1024),
		WithWaitBeforeDropDB(0),
		WithOpenedCacheSize(1000),
		WithMoveBatchSize(3))
	require.NoError(t, b.Open(mode.ComponentReadWrite))
	require.NoError(t, b.Init())

	eg, egCtx := errgroup.WithContext(context.Background())
	storageIDs := make(map[oid.Address][]byte)
	storageIDsGuard := &sync.Mutex{}
	for range 100 {
		eg.Go(func() error {
			obj := blobstortest.NewObject(1024)
			data, err := obj.Marshal()
			if err != nil {
				return err
			}
			var prm common.PutPrm
			prm.Address = object.AddressOf(obj)
			prm.RawData = data
			res, err := b.Put(egCtx, prm)
			if err != nil {
				return err
			}
			storageIDsGuard.Lock()
			storageIDs[prm.Address] = res.StorageID
			storageIDsGuard.Unlock()
			return nil
		})
	}

	require.NoError(t, eg.Wait())
	require.NoError(t, b.Close(context.Background()))

	b = NewBlobovniczaTree(
		context.Background(),
		WithLogger(test.NewLogger(t)),
		WithObjectSizeLimit(2048),
		WithBlobovniczaShallowWidth(targetWidth),
		WithBlobovniczaShallowDepth(targetDepth),
		WithRootPath(dir),
		WithBlobovniczaSize(100*1024),
		WithWaitBeforeDropDB(0),
		WithOpenedCacheSize(1000),
		WithMoveBatchSize(50))
	require.NoError(t, b.Open(mode.ComponentReadWrite))
	require.NoError(t, b.Init())

	for addr, storageID := range storageIDs {
		var gPrm common.GetPrm
		gPrm.Address = addr
		gPrm.StorageID = storageID
		_, err := b.Get(context.Background(), gPrm)
		require.NoError(t, err)
	}

	metaStub := &storageIDUpdateStub{
		storageIDs: storageIDs,
		guard:      &sync.Mutex{},
	}
	var rPrm common.RebuildPrm
	rPrm.MetaStorage = metaStub
	rPrm.WorkerLimiter = &rebuildLimiterStub{}
	rPrm.FillPercent = 1
	rRes, err := b.Rebuild(context.Background(), rPrm)
	require.NoError(t, err)
	dataMigrated := rRes.ObjectsMoved > 0 || rRes.FilesRemoved > 0 || metaStub.updatedCount > 0
	require.Equal(t, shouldMigrate, dataMigrated)

	for addr, storageID := range storageIDs {
		var gPrm common.GetPrm
		gPrm.Address = addr
		gPrm.StorageID = storageID
		_, err := b.Get(context.Background(), gPrm)
		require.NoError(t, err)
	}

	require.NoError(t, b.Close(context.Background()))
}

type storageIDUpdateStub struct {
	guard        *sync.Mutex
	storageIDs   map[oid.Address][]byte
	updatedCount uint64
}

func (s *storageIDUpdateStub) UpdateStorageID(ctx context.Context, addr oid.Address, storageID []byte) error {
	s.guard.Lock()
	defer s.guard.Unlock()

	s.storageIDs[addr] = storageID
	s.updatedCount++
	return nil
}

type rebuildLimiterStub struct{}

func (s *rebuildLimiterStub) AcquireWorkSlot(context.Context) error { return nil }
func (s *rebuildLimiterStub) ReleaseWorkSlot()                      {}
