//go:build linux && !fstree_generic

package fstree

import (
	"errors"
	"io/fs"
	"strconv"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/util/logicerr"
	utilSync "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/sync"
	apistatus "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client/status"
	"golang.org/x/sys/unix"
)

type linuxWriter struct {
	root  string
	perm  uint32
	flags int

	fileGuard          keyLock
	fileCounter        FileCounter
	fileCounterEnabled bool
}

func newSpecificWriteData(c FileCounter, root string, perm fs.FileMode, noSync bool) writer {
	flags := unix.O_WRONLY | unix.O_TMPFILE | unix.O_CLOEXEC
	if !noSync {
		flags |= unix.O_DSYNC
	}
	fd, err := unix.Open(root, flags, uint32(perm))
	if err != nil {
		// Which means that OS-specific writeData can't be created
		// and FSTree should use the generic one.
		return nil
	}
	_ = unix.Close(fd) // Don't care about error.
	var fileGuard keyLock = &noopKeyLock{}
	fileCounterEnabled := counterEnabled(c)
	if fileCounterEnabled {
		fileGuard = utilSync.NewKeyLocker[string]()
	}
	w := &linuxWriter{
		root:               root,
		perm:               uint32(perm),
		flags:              flags,
		fileGuard:          fileGuard,
		fileCounter:        c,
		fileCounterEnabled: fileCounterEnabled,
	}
	return w
}

func (w *linuxWriter) writeData(p string, data []byte) error {
	err := w.writeFile(p, data)
	if errors.Is(err, unix.ENOSPC) {
		return common.ErrNoSpace
	}
	return err
}

func (w *linuxWriter) writeFile(p string, data []byte) error {
	if w.fileCounterEnabled {
		w.fileGuard.Lock(p)
		defer w.fileGuard.Unlock(p)
	}
	fd, err := unix.Open(w.root, w.flags, w.perm)
	if err != nil {
		return err
	}
	written := 0
	tmpPath := "/proc/self/fd/" + strconv.FormatUint(uint64(fd), 10)
	n, err := unix.Write(fd, data)
	for err == nil {
		written += n

		if written == len(data) {
			err = unix.Linkat(unix.AT_FDCWD, tmpPath, unix.AT_FDCWD, p, unix.AT_SYMLINK_FOLLOW)
			if err == nil {
				w.fileCounter.Inc(uint64(len(data)))
			}
			if errors.Is(err, unix.EEXIST) {
				err = nil
			}
			break
		}

		// From man 2 write:
		// https://www.man7.org/linux/man-pages/man2/write.2.html
		//
		// Note that a successful write() may transfer fewer than count
		// bytes.  Such partial writes can occur for various reasons; for
		// example, because there was insufficient space on the disk device
		// to write all of the requested bytes, or because a blocked write()
		// to a socket, pipe, or similar was interrupted by a signal handler
		// after it had transferred some, but before it had transferred all
		// of the requested bytes.  In the event of a partial write, the
		// caller can make another write() call to transfer the remaining
		// bytes.  The subsequent call will either transfer further bytes or
		// may result in an error (e.g., if the disk is now full).
		n, err = unix.Write(fd, data[written:])
	}
	errClose := unix.Close(fd)
	if err != nil {
		return err // Close() error is ignored, we have a better one.
	}
	return errClose
}

func (w *linuxWriter) removeFile(p string, size uint64) error {
	if w.fileCounterEnabled {
		w.fileGuard.Lock(p)
		defer w.fileGuard.Unlock(p)

		if size == 0 {
			var stat unix.Stat_t
			err := unix.Stat(p, &stat)
			if err != nil {
				if err == unix.ENOENT {
					return logicerr.Wrap(new(apistatus.ObjectNotFound))
				}
				return err
			}
			size = uint64(stat.Size)
		}
	}

	err := unix.Unlink(p)
	if err != nil && err == unix.ENOENT {
		return logicerr.Wrap(new(apistatus.ObjectNotFound))
	}
	if err == nil {
		w.fileCounter.Dec(size)
	}
	return err
}
