package meta_test

import (
	"context"
	"errors"
	"strconv"
	"testing"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/object"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/internal/testutil"
	meta "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/metabase"
	objectV2 "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/api/object"
	cidSDK "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/id"
	cidtest "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/id/test"
	objectSDK "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	oidtest "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id/test"
	"github.com/stretchr/testify/require"
	"go.etcd.io/bbolt"
)

func BenchmarkListWithCursor(b *testing.B) {
	db := listWithCursorPrepareDB(b)
	defer func() { require.NoError(b, db.Close(context.Background())) }()

	b.Run("1 item", func(b *testing.B) {
		benchmarkListWithCursor(b, db, 1)
	})
	b.Run("10 items", func(b *testing.B) {
		benchmarkListWithCursor(b, db, 10)
	})
	b.Run("100 items", func(b *testing.B) {
		benchmarkListWithCursor(b, db, 100)
	})
}

func listWithCursorPrepareDB(b *testing.B) *meta.DB {
	db := newDB(b, meta.WithMaxBatchSize(1), meta.WithBoltDBOptions(&bbolt.Options{
		NoSync: true,
	})) // faster single-thread generation

	obj := testutil.GenerateObject()
	for i := range 100_000 { // should be a multiple of all batch sizes
		obj.SetID(oidtest.ID())
		if i%9 == 0 { // let's have 9 objects per container
			obj.SetContainerID(cidtest.ID())
		}
		require.NoError(b, putBig(db, obj))
	}
	return db
}

func benchmarkListWithCursor(b *testing.B, db *meta.DB, batchSize int) {
	var prm meta.ListPrm
	prm.SetCount(uint32(batchSize))

	b.ResetTimer()
	b.ReportAllocs()
	for range b.N {
		res, err := db.ListWithCursor(context.Background(), prm)
		if err != nil {
			if errors.Is(err, meta.ErrEndOfListing) {
				b.Fatalf("error: %v", err)
			}
			prm.SetCursor(nil)
		} else if ln := len(res.AddressList()); ln != batchSize {
			b.Fatalf("invalid batch size: %d", ln)
		} else {
			prm.SetCursor(res.Cursor())
		}
	}
}

func TestLisObjectsWithCursor(t *testing.T) {
	t.Parallel()

	const (
		currEpoch  = 100
		expEpoch   = currEpoch - 1
		containers = 5
		total      = containers * 6 // regular + ts + child + lock + non-expired regular + locked expired
	)

	db := newDB(t, meta.WithEpochState(epochState{currEpoch}))
	defer func() { require.NoError(t, db.Close(context.Background())) }()

	expected := make([]object.Info, 0, total)

	// fill metabase with objects
	for range containers {
		containerID := cidtest.ID()

		// add one regular object
		obj := testutil.GenerateObjectWithCID(containerID)
		obj.SetType(objectSDK.TypeRegular)
		err := putBig(db, obj)
		require.NoError(t, err)
		expected = append(expected, object.Info{Address: object.AddressOf(obj), Type: objectSDK.TypeRegular})

		// add one tombstone
		obj = testutil.GenerateObjectWithCID(containerID)
		obj.SetType(objectSDK.TypeTombstone)
		err = putBig(db, obj)
		require.NoError(t, err)
		expected = append(expected, object.Info{Address: object.AddressOf(obj), Type: objectSDK.TypeTombstone})

		// add one lock
		obj = testutil.GenerateObjectWithCID(containerID)
		obj.SetType(objectSDK.TypeLock)
		err = putBig(db, obj)
		require.NoError(t, err)
		expected = append(expected, object.Info{Address: object.AddressOf(obj), Type: objectSDK.TypeLock})

		// add one inhumed (do not include into expected)
		obj = testutil.GenerateObjectWithCID(containerID)
		obj.SetType(objectSDK.TypeRegular)
		err = putBig(db, obj)
		require.NoError(t, err)
		ts := testutil.GenerateObjectWithCID(containerID)
		err = metaInhume(db, object.AddressOf(obj), object.AddressOf(ts).Object())
		require.NoError(t, err)

		// add one child object (do not include parent into expected)
		splitID := objectSDK.NewSplitID()
		parent := testutil.GenerateObjectWithCID(containerID)
		testutil.AddAttribute(parent, "foo", "bar")
		child := testutil.GenerateObjectWithCID(containerID)
		child.SetParent(parent)
		idParent, _ := parent.ID()
		child.SetParentID(idParent)
		child.SetSplitID(splitID)
		err = putBig(db, child)
		require.NoError(t, err)
		expected = append(expected, object.Info{Address: object.AddressOf(child), Type: objectSDK.TypeRegular})

		// add expired object (do not include into expected)
		obj = testutil.GenerateObjectWithCID(containerID)
		testutil.AddAttribute(obj, objectV2.SysAttributeExpEpoch, strconv.Itoa(expEpoch))
		require.NoError(t, metaPut(db, obj, nil))

		// add non-expired object (include into expected)
		obj = testutil.GenerateObjectWithCID(containerID)
		testutil.AddAttribute(obj, objectV2.SysAttributeExpEpoch, strconv.Itoa(currEpoch))
		require.NoError(t, metaPut(db, obj, nil))
		expected = append(expected, object.Info{Address: object.AddressOf(obj), Type: objectSDK.TypeRegular})

		// add locked expired object (include into expected)
		obj = testutil.GenerateObjectWithCID(containerID)
		objID := oidtest.ID()
		obj.SetID(objID)
		testutil.AddAttribute(obj, objectV2.SysAttributeExpEpoch, strconv.Itoa(expEpoch))
		require.NoError(t, metaPut(db, obj, nil))
		require.NoError(t, db.Lock(context.Background(), containerID, oidtest.ID(), []oid.ID{objID}))
		expected = append(expected, object.Info{Address: object.AddressOf(obj), Type: objectSDK.TypeRegular})
	}

	t.Run("success with various count", func(t *testing.T) {
		for countPerReq := 1; countPerReq <= total; countPerReq++ {
			got := make([]object.Info, 0, total)

			res, cursor, err := metaListWithCursor(db, uint32(countPerReq), nil)
			require.NoError(t, err, "count:%d", countPerReq)
			got = append(got, res...)

			expectedIterations := total / countPerReq
			if total%countPerReq == 0 { // remove initial list if aligned
				expectedIterations--
			}

			for range expectedIterations {
				res, cursor, err = metaListWithCursor(db, uint32(countPerReq), cursor)
				require.NoError(t, err, "count:%d", countPerReq)
				got = append(got, res...)
			}

			_, _, err = metaListWithCursor(db, uint32(countPerReq), cursor)
			require.ErrorIs(t, err, meta.ErrEndOfListing, "count:%d", countPerReq, cursor)
			require.ElementsMatch(t, expected, got, "count:%d", countPerReq)
		}
	})

	t.Run("invalid count", func(t *testing.T) {
		_, _, err := metaListWithCursor(db, 0, nil)
		require.ErrorIs(t, err, meta.ErrEndOfListing)
	})
}

func TestAddObjectDuringListingWithCursor(t *testing.T) {
	t.Parallel()

	db := newDB(t)
	defer func() { require.NoError(t, db.Close(context.Background())) }()

	const total = 5

	expected := make(map[string]int, total)

	// fill metabase with objects
	for range total {
		obj := testutil.GenerateObject()
		err := putBig(db, obj)
		require.NoError(t, err)
		expected[object.AddressOf(obj).EncodeToString()] = 0
	}

	// get half of the objects
	got, cursor, err := metaListWithCursor(db, total/2, nil)
	require.NoError(t, err)
	for _, obj := range got {
		if _, ok := expected[obj.Address.EncodeToString()]; ok {
			expected[obj.Address.EncodeToString()]++
		}
	}

	// add new objects
	for range total {
		obj := testutil.GenerateObject()
		err = putBig(db, obj)
		require.NoError(t, err)
	}

	// get remaining objects
	for {
		got, cursor, err = metaListWithCursor(db, total, cursor)
		if errors.Is(err, meta.ErrEndOfListing) {
			break
		}
		for _, obj := range got {
			if _, ok := expected[obj.Address.EncodeToString()]; ok {
				expected[obj.Address.EncodeToString()]++
			}
		}
	}

	// check if all expected objects were fetched after database update
	for _, v := range expected {
		require.Equal(t, 1, v)
	}
}

func metaListWithCursor(db *meta.DB, count uint32, cursor *meta.Cursor) ([]object.Info, *meta.Cursor, error) {
	var listPrm meta.ListPrm
	listPrm.SetCount(count)
	listPrm.SetCursor(cursor)

	r, err := db.ListWithCursor(context.Background(), listPrm)
	return r.AddressList(), r.Cursor(), err
}

func TestIterateOver(t *testing.T) {
	t.Parallel()

	db := newDB(t)
	defer func() { require.NoError(t, db.Close(context.Background())) }()

	const total uint64 = 5
	for _, typ := range []objectSDK.Type{objectSDK.TypeRegular, objectSDK.TypeTombstone, objectSDK.TypeLock} {
		var expected []*objectSDK.Object
		// fill metabase with objects
		cid := cidtest.ID()
		for range total {
			obj := testutil.GenerateObjectWithCID(cid)
			obj.SetType(typ)
			err := metaPut(db, obj, nil)
			require.NoError(t, err)
			expected = append(expected, obj)
		}

		var metaIter meta.IterateOverObjectsInContainerPrm
		var count uint64
		metaIter.Handler = func(context.Context, *object.Info) error {
			count++
			return nil
		}
		metaIter.ContainerID = cid
		metaIter.ObjectType = typ
		err := db.IterateOverObjectsInContainer(context.Background(), metaIter)
		require.NoError(t, err)
		require.Equal(t, total, count)

		var metaCount meta.CountAliveObjectsInContainerPrm
		metaCount.ContainerID = cid
		metaCount.ObjectType = typ
		res, err := db.CountAliveObjectsInContainer(context.Background(), metaCount)
		require.NoError(t, err)
		require.Equal(t, res, total)

		err = metaDelete(db, object.AddressOf(expected[0]), object.AddressOf(expected[1]))
		require.NoError(t, err)

		res, err = db.CountAliveObjectsInContainer(context.Background(), metaCount)
		require.NoError(t, err)
		require.Equal(t, uint64(3), res)
	}
	var count int
	var metaPrm meta.IterateOverContainersPrm
	metaPrm.Handler = func(context.Context, objectSDK.Type, cidSDK.ID) error {
		count++
		return nil
	}
	err := db.IterateOverContainers(context.Background(), metaPrm)
	require.NoError(t, err)
	require.Equal(t, 3, count)
}
