package shard

import (
	"context"
	"errors"
	"fmt"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/shard/mode"
	"github.com/mr-tron/base58"
	"go.uber.org/zap"
)

// ID represents Shard identifier.
//
// Each shard should have the unique ID within
// a single instance of local storage.
type ID []byte

// NewIDFromBytes constructs ID from byte slice.
func NewIDFromBytes(v []byte) *ID {
	return (*ID)(&v)
}

func (id ID) String() string {
	return base58.Encode(id)
}

// ID returns Shard identifier.
func (s *Shard) ID() *ID {
	return s.info.ID
}

// UpdateID reads shard ID saved in the metabase and updates it if it is missing.
func (s *Shard) UpdateID(ctx context.Context) (err error) {
	var idFromMetabase []byte
	modeDegraded := s.GetMode().NoMetabase()
	if !modeDegraded {
		if idFromMetabase, err = s.metaBase.GetShardID(ctx, mode.ReadOnly); err != nil {
			err = fmt.Errorf("read shard id from metabase: %w", err)
		}
	}

	if len(idFromMetabase) != 0 {
		s.info.ID = NewIDFromBytes(idFromMetabase)
	}

	shardID := s.info.ID.String()
	s.cfg.metricsWriter.SetShardID(shardID)
	if s.writeCache != nil && s.writeCache.GetMetrics() != nil {
		s.writeCache.GetMetrics().SetShardID(shardID)
	}

	s.log = s.log.With(zap.Stringer("shard_id", s.info.ID))
	s.metaBase.SetLogger(s.log)
	s.blobStor.SetLogger(s.log)
	if s.hasWriteCache() {
		s.writeCache.SetLogger(s.log)
	}
	s.metaBase.SetParentID(s.info.ID.String())
	s.blobStor.SetParentID(s.info.ID.String())
	if s.pilorama != nil {
		s.pilorama.SetParentID(s.info.ID.String())
	}

	if len(idFromMetabase) == 0 && !modeDegraded {
		if setErr := s.metaBase.SetShardID(ctx, *s.info.ID, s.GetMode()); setErr != nil {
			err = errors.Join(err, fmt.Errorf("write shard id to metabase: %w", setErr))
		}
	}
	return
}
