package event

import (
	"context"
	"errors"
	"fmt"
	"sync"

	"git.frostfs.info/TrueCloudLab/frostfs-node/internal/logs"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/morph/client"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/morph/subscriber"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/logger"
	"github.com/nspcc-dev/neo-go/pkg/core/block"
	"github.com/nspcc-dev/neo-go/pkg/core/state"
	"github.com/nspcc-dev/neo-go/pkg/neorpc/result"
	"github.com/nspcc-dev/neo-go/pkg/util"
	"github.com/panjf2000/ants/v2"
	"go.uber.org/zap"
)

// Listener is an interface of smart contract notification event listener.
type Listener interface {
	// Listen must start the event listener.
	//
	// Must listen to events with the parser installed.
	Listen(context.Context)

	// ListenWithError must start the event listener.
	//
	// Must listen to events with the parser installed.
	//
	// Must send error to channel if subscriber channel has been closed or
	// it could not be started.
	ListenWithError(context.Context, chan<- error)

	// RegisterNotificationHandler must register the event handler for particular notification event of contract.
	//
	// The specified handler must be called after each capture and parsing of the event.
	//
	// Must ignore nil handlers.
	RegisterNotificationHandler(NotificationHandlerInfo)

	// EnableNotarySupport enables notary request listening. Passed hash is
	// notary mainTX signer. In practise, it means that listener will subscribe
	// for only notary requests that are going to be paid with passed hash.
	//
	// Must not be called after Listen or ListenWithError.
	EnableNotarySupport(util.Uint160, client.AlphabetKeys, BlockCounter)

	// SetNotaryParser must set the parser of particular notary request event.
	//
	// Parser of each event must be set once. All parsers must be set before Listen call.
	//
	// Must ignore nil parsers and all calls after listener has been started.
	//
	// Has no effect if EnableNotarySupport was not called before Listen or ListenWithError.
	SetNotaryParser(NotaryParserInfo)

	// RegisterNotaryHandler must register the event handler for particular notification event of contract.
	//
	// The specified handler must be called after each capture and parsing of the event.
	//
	// Must ignore nil handlers.
	//
	// Has no effect if EnableNotarySupport was not called before Listen or ListenWithError.
	RegisterNotaryHandler(NotaryHandlerInfo)

	// RegisterBlockHandler must register chain block handler.
	//
	// The specified handler must be called after each capture and parsing of the new block from chain.
	//
	// Must ignore nil handlers.
	RegisterBlockHandler(BlockHandler)

	// Stop must stop the event listener.
	Stop()
}

// ListenerParams is a group of parameters
// for Listener constructor.
type ListenerParams struct {
	Logger *logger.Logger

	Subscriber subscriber.Subscriber

	WorkerPoolCapacity int
}

type listener struct {
	mtx sync.RWMutex

	wg sync.WaitGroup

	startOnce, stopOnce sync.Once

	notificationParsers  map[scriptHashWithType]NotificationParser
	notificationHandlers map[scriptHashWithType][]Handler

	listenNotary           bool
	notaryEventsPreparator NotaryPreparator
	notaryParsers          map[notaryRequestTypes]NotaryParser
	notaryHandlers         map[notaryRequestTypes]Handler
	notaryMainTXSigner     util.Uint160 // filter for notary subscription

	log *logger.Logger

	subscriber subscriber.Subscriber

	blockHandlers []BlockHandler

	pool *ants.Pool
}

const newListenerFailMsg = "instantiate Listener"

var (
	errNilLogger = errors.New("nil logger")

	errNilSubscriber = errors.New("nil event subscriber")

	errNotificationSubscrConnectionTerminated = errors.New("event subscriber connection has been terminated")

	errNotarySubscrConnectionTerminated = errors.New("notary event subscriber connection has been terminated")

	errBlockNotificationChannelClosed = errors.New("new block notification channel is closed")
)

// Listen starts the listening for events with registered handlers.
//
// Executes once, all subsequent calls do nothing.
//
// Returns an error if listener was already started.
func (l *listener) Listen(ctx context.Context) {
	l.startOnce.Do(func() {
		l.wg.Add(1)
		defer l.wg.Done()

		l.listen(ctx, nil)
	})
}

// ListenWithError starts the listening for events with registered handlers and
// passing error message to intError channel if subscriber channel has been closed.
//
// Executes once, all subsequent calls do nothing.
//
// Returns an error if listener was already started.
func (l *listener) ListenWithError(ctx context.Context, intError chan<- error) {
	l.startOnce.Do(func() {
		l.wg.Add(1)
		defer l.wg.Done()

		l.listen(ctx, intError)
	})
}

func (l *listener) listen(ctx context.Context, intError chan<- error) {
	subErrCh := make(chan error)

	go l.subscribe(subErrCh)

	l.listenLoop(ctx, intError, subErrCh)
}

func (l *listener) subscribe(errCh chan error) {
	l.wg.Add(1)
	defer l.wg.Done()
	// create the list of listening contract hashes
	hashes := make([]util.Uint160, 0)

	// fill the list with the contracts with set event parsers.
	l.mtx.RLock()
	for hashType := range l.notificationParsers {
		scHash := hashType.Hash

		// prevent repetitions
		for _, hash := range hashes {
			if hash.Equals(scHash) {
				continue
			}
		}

		hashes = append(hashes, hashType.Hash)
	}
	l.mtx.RUnlock()

	err := l.subscriber.SubscribeForNotification(hashes...)
	if err != nil {
		errCh <- fmt.Errorf("subscribe for notifications: %w", err)
		return
	}

	if len(l.blockHandlers) > 0 {
		if err = l.subscriber.BlockNotifications(); err != nil {
			errCh <- fmt.Errorf("subscribe for blocks: %w", err)
			return
		}
	}

	if l.listenNotary {
		if err = l.subscriber.SubscribeForNotaryRequests(l.notaryMainTXSigner); err != nil {
			errCh <- fmt.Errorf("subscribe for notary requests: %w", err)
			return
		}
	}
}

func (l *listener) sendError(ctx context.Context, intErr chan<- error, err error) bool {
	if intErr == nil {
		return false
	}
	// This select required because were are reading from error channel and closing listener
	// in the same routine when shutting down node.
	select {
	case <-ctx.Done():
		l.log.Info(ctx, logs.EventStopEventListenerByContext,
			zap.String("reason", ctx.Err().Error()),
		)
		return false
	case intErr <- err:
		return true
	}
}

func (l *listener) listenLoop(ctx context.Context, intErr chan<- error, subErrCh chan error) {
	chs := l.subscriber.NotificationChannels()

loop:
	for {
		select {
		case err := <-subErrCh:
			if !l.sendError(ctx, intErr, err) {
				l.log.Error(ctx, logs.EventStopEventListenerByError, zap.Error(err))
			}
			break loop
		case <-ctx.Done():
			l.log.Info(ctx, logs.EventStopEventListenerByContext,
				zap.String("reason", ctx.Err().Error()),
			)
			break loop
		case notifyEvent, ok := <-chs.NotificationsCh:
			if !ok {
				l.log.Warn(ctx, logs.EventStopEventListenerByNotificationChannel)
				l.sendError(ctx, intErr, errNotificationSubscrConnectionTerminated)
				break loop
			} else if notifyEvent == nil {
				l.log.Warn(ctx, logs.EventNilNotificationEventWasCaught)
				continue loop
			}

			l.handleNotifyEvent(ctx, notifyEvent)
		case notaryEvent, ok := <-chs.NotaryRequestsCh:
			if !ok {
				l.log.Warn(ctx, logs.EventStopEventListenerByNotaryChannel)
				l.sendError(ctx, intErr, errNotarySubscrConnectionTerminated)
				break loop
			} else if notaryEvent == nil {
				l.log.Warn(ctx, logs.EventNilNotaryEventWasCaught)
				continue loop
			}

			l.handleNotaryEvent(ctx, notaryEvent)
		case b, ok := <-chs.BlockCh:
			if !ok {
				l.log.Warn(ctx, logs.EventStopEventListenerByBlockChannel)
				l.sendError(ctx, intErr, errBlockNotificationChannelClosed)
				break loop
			} else if b == nil {
				l.log.Warn(ctx, logs.EventNilBlockWasCaught)
				continue loop
			}

			l.handleBlockEvent(ctx, b)
		}
	}
}

func (l *listener) handleBlockEvent(ctx context.Context, b *block.Block) {
	if err := l.pool.Submit(func() {
		for i := range l.blockHandlers {
			l.blockHandlers[i](ctx, b)
		}
	}); err != nil {
		l.log.Warn(ctx, logs.EventListenerWorkerPoolDrained,
			zap.Int("capacity", l.pool.Cap()))
	}
}

func (l *listener) handleNotaryEvent(ctx context.Context, notaryEvent *result.NotaryRequestEvent) {
	if err := l.pool.Submit(func() {
		l.parseAndHandleNotary(ctx, notaryEvent)
	}); err != nil {
		l.log.Warn(ctx, logs.EventListenerWorkerPoolDrained,
			zap.Int("capacity", l.pool.Cap()))
	}
}

func (l *listener) handleNotifyEvent(ctx context.Context, notifyEvent *state.ContainedNotificationEvent) {
	if err := l.pool.Submit(func() {
		l.parseAndHandleNotification(ctx, notifyEvent)
	}); err != nil {
		l.log.Warn(ctx, logs.EventListenerWorkerPoolDrained,
			zap.Int("capacity", l.pool.Cap()))
	}
}

func (l *listener) parseAndHandleNotification(ctx context.Context, notifyEvent *state.ContainedNotificationEvent) {
	log := l.log.With(
		zap.String("script hash LE", notifyEvent.ScriptHash.StringLE()),
	)

	// calculate event type from bytes
	typEvent := TypeFromString(notifyEvent.Name)

	log = log.With(
		zap.String("event type", notifyEvent.Name),
	)

	// get the event parser
	keyEvent := scriptHashWithType{Hash: notifyEvent.ScriptHash, Type: typEvent}

	l.mtx.RLock()
	parser, ok := l.notificationParsers[keyEvent]
	l.mtx.RUnlock()

	if !ok {
		log.Debug(ctx, logs.EventEventParserNotSet)

		return
	}

	// parse the notification event
	event, err := parser(notifyEvent)
	if err != nil {
		log.Warn(ctx, logs.EventCouldNotParseNotificationEvent,
			zap.Error(err),
		)

		return
	}

	// handler the event
	l.mtx.RLock()
	handlers := l.notificationHandlers[keyEvent]
	l.mtx.RUnlock()

	if len(handlers) == 0 {
		log.Info(ctx, logs.EventNotificationHandlersForParsedNotificationEventWereNotRegistered,
			zap.Any("event", event),
		)

		return
	}

	for _, handler := range handlers {
		handler(ctx, event)
	}
}

func (l *listener) parseAndHandleNotary(ctx context.Context, nr *result.NotaryRequestEvent) {
	// prepare the notary event
	notaryEvent, err := l.notaryEventsPreparator.Prepare(nr.NotaryRequest)
	if err != nil {
		var expErr *ExpiredTXError
		switch {
		case errors.Is(err, ErrTXAlreadyHandled):
		case errors.As(err, &expErr):
			l.log.Warn(ctx, logs.EventSkipExpiredMainTXNotaryEvent,
				zap.Error(err),
				zap.Uint32("current_block_height", expErr.CurrentBlockHeight),
				zap.Uint32("fallback_tx_not_valid_before_height", expErr.FallbackTXNotValidBeforeHeight),
			)
		default:
			l.log.Warn(ctx, logs.EventCouldNotPrepareAndValidateNotaryEvent,
				zap.Error(err),
			)
		}

		return
	}

	log := l.log.With(
		zap.String("contract", notaryEvent.ScriptHash().StringLE()),
		zap.Stringer("method", notaryEvent.Type()),
	)

	notaryKey := notaryRequestTypes{}
	notaryKey.SetMempoolType(nr.Type)
	notaryKey.SetRequestType(notaryEvent.Type())
	notaryKey.SetScriptHash(notaryEvent.ScriptHash())

	// get notary parser
	l.mtx.RLock()
	parser, ok := l.notaryParsers[notaryKey]
	l.mtx.RUnlock()

	if !ok {
		log.Debug(ctx, logs.EventNotaryParserNotSet)

		return
	}

	// parse the notary event
	event, err := parser(notaryEvent)
	if err != nil {
		log.Warn(ctx, logs.EventCouldNotParseNotaryEvent,
			zap.Error(err),
		)

		return
	}

	// handle the event
	l.mtx.RLock()
	handler, ok := l.notaryHandlers[notaryKey]
	l.mtx.RUnlock()

	if !ok {
		log.Info(ctx, logs.EventNotaryHandlersForParsedNotificationEventWereNotRegistered,
			zap.Any("event", event),
		)

		return
	}

	handler(ctx, event)
}

// RegisterNotificationHandler registers the handler for particular notification event of contract.
//
// Ignores nil handlers.
// Ignores handlers of event without parser.
func (l *listener) RegisterNotificationHandler(hi NotificationHandlerInfo) {
	log := l.log.With(
		zap.String("contract", hi.Contract.StringLE()),
		zap.Stringer("event_type", hi.Type),
	)

	// check if parser was set
	l.mtx.Lock()
	defer l.mtx.Unlock()

	k := scriptHashWithType{Hash: hi.Contract, Type: hi.Type}

	l.notificationParsers[k] = hi.Parser
	l.notificationHandlers[k] = append(
		l.notificationHandlers[k],
		hi.Handlers...,
	)

	log.Debug(context.Background(), logs.EventRegisteredNewEventHandler)
}

// EnableNotarySupport enables notary request listening. Passed hash is
// notary mainTX signer. In practise, it means that listener will subscribe
// for only notary requests that are going to be paid with passed hash.
//
// Must not be called after Listen or ListenWithError.
func (l *listener) EnableNotarySupport(mainTXSigner util.Uint160, alphaKeys client.AlphabetKeys, bc BlockCounter) {
	l.mtx.Lock()
	defer l.mtx.Unlock()

	l.listenNotary = true
	l.notaryMainTXSigner = mainTXSigner
	l.notaryHandlers = make(map[notaryRequestTypes]Handler)
	l.notaryParsers = make(map[notaryRequestTypes]NotaryParser)
	l.notaryEventsPreparator = notaryPreparator(
		PreparatorPrm{
			AlphaKeys:    alphaKeys,
			BlockCounter: bc,
		},
	)
}

// SetNotaryParser sets the parser of particular notary request event.
//
// Ignores nil and already set parsers.
// Ignores the parser if listener is started.
func (l *listener) SetNotaryParser(pi NotaryParserInfo) {
	if !l.listenNotary {
		return
	}

	log := l.log.With(
		zap.Stringer("mempool_type", pi.GetMempoolType()),
		zap.String("contract", pi.ScriptHash().StringLE()),
		zap.Stringer("notary_type", pi.RequestType()),
	)

	l.mtx.Lock()
	defer l.mtx.Unlock()

	// add event parser
	if _, ok := l.notaryParsers[pi.notaryRequestTypes]; !ok {
		l.notaryParsers[pi.notaryRequestTypes] = pi.parser()
	}

	log.Info(context.Background(), logs.EventRegisteredNewEventParser)
}

// RegisterNotaryHandler registers the handler for particular notification notary request event.
//
// Ignores nil handlers.
// Ignores handlers of event without parser.
func (l *listener) RegisterNotaryHandler(hi NotaryHandlerInfo) {
	if !l.listenNotary {
		return
	}

	log := l.log.With(
		zap.Stringer("mempool_type", hi.GetMempoolType()),
		zap.String("contract", hi.ScriptHash().StringLE()),
		zap.Stringer("notary type", hi.RequestType()),
	)

	// check if parser was set
	l.mtx.RLock()
	_, ok := l.notaryParsers[hi.notaryRequestTypes]
	l.mtx.RUnlock()

	if !ok {
		log.Warn(context.Background(), logs.EventIgnoreHandlerOfNotaryEventWoParser)
		return
	}

	// add notary event handler
	l.mtx.Lock()
	l.notaryHandlers[hi.notaryRequestTypes] = hi.Handler()
	l.mtx.Unlock()

	log.Info(context.Background(), logs.EventRegisteredNewEventHandler)
}

// Stop closes subscription channel with remote neo node.
func (l *listener) Stop() {
	l.stopOnce.Do(func() {
		l.subscriber.Close()
		l.pool.Release()
	})
	l.wg.Wait()
}

func (l *listener) RegisterBlockHandler(handler BlockHandler) {
	l.blockHandlers = append(l.blockHandlers, handler)
}

// NewListener create the notification event listener instance and returns Listener interface.
func NewListener(p ListenerParams) (Listener, error) {
	switch {
	case p.Logger == nil:
		return nil, fmt.Errorf("%s: %w", newListenerFailMsg, errNilLogger)
	case p.Subscriber == nil:
		return nil, fmt.Errorf("%s: %w", newListenerFailMsg, errNilSubscriber)
	}

	// The pool here must be blocking, otherwise notifications could be dropped.
	// The default capacity is 0, which means "infinite".
	pool, err := ants.NewPool(p.WorkerPoolCapacity)
	if err != nil {
		return nil, fmt.Errorf("init worker pool: %w", err)
	}

	return &listener{
		notificationParsers:  make(map[scriptHashWithType]NotificationParser),
		notificationHandlers: make(map[scriptHashWithType][]Handler),
		log:                  p.Logger,
		subscriber:           p.Subscriber,
		pool:                 pool,
	}, nil
}
