package getsvc

import (
	"context"
	"errors"

	"git.frostfs.info/TrueCloudLab/frostfs-node/internal/logs"
	apistatus "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client/status"
	objectSDK "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object"
	"go.uber.org/zap"
)

func (r *request) assembleEC(ctx context.Context) {
	if r.isRaw() {
		r.log.Debug(ctx, logs.GetCanNotAssembleTheObject)
		return
	}

	// Any access tokens are not expected to be used in the assembly process:
	//  - there is no requirement to specify child objects in session/bearer
	//    token for `GET`/`GETRANGE`/`RANGEHASH` requests in the API protocol,
	//    and, therefore, their missing in the original request should not be
	//    considered as error; on the other hand, without session for every child
	//    object, it is impossible to attach bearer token in the new generated
	//    requests correctly because the token has not been issued for that node's
	//    key;
	//  - the assembly process is expected to be handled on a container node
	//    only since the requests forwarding mechanism presentation; such the
	//    node should have enough rights for getting any child object by design.
	r.prm.common.ForgetTokens()

	// Do not use forwarding during assembly stage.
	// Request forwarding closure inherited in produced
	// `execCtx` so it should be disabled there.
	r.disableForwarding()

	r.log.Debug(ctx, logs.GetTryingToAssembleTheECObject)

	// initialize epoch number
	ok := r.initEpoch(ctx)
	if !ok {
		return
	}

	r.prm.common = r.prm.common.WithLocalOnly(false)
	assembler := newAssemblerEC(r.address(), r.infoEC, r.ctxRange(), r, r.localStorage, r.log, r.headOnly(), r.traverserGenerator, r.curProcEpoch)

	r.log.Debug(ctx, logs.GetAssemblingECObject,
		zap.Uint64("range_offset", r.ctxRange().GetOffset()),
		zap.Uint64("range_length", r.ctxRange().GetLength()),
	)
	defer r.log.Debug(ctx, logs.GetAssemblingECObjectCompleted,
		zap.Uint64("range_offset", r.ctxRange().GetOffset()),
		zap.Uint64("range_length", r.ctxRange().GetLength()),
	)

	obj, err := assembler.Assemble(ctx, r.prm.objWriter)
	if err != nil && !errors.As(err, new(*objectSDK.ECInfoError)) {
		r.log.Warn(ctx, logs.GetFailedToAssembleECObject,
			zap.Error(err),
			zap.Uint64("range_offset", r.ctxRange().GetOffset()),
			zap.Uint64("range_length", r.ctxRange().GetLength()),
		)
	}

	var errRemoved *apistatus.ObjectAlreadyRemoved
	var errOutOfRange *apistatus.ObjectOutOfRange
	var errECInfo *objectSDK.ECInfoError

	switch {
	default:
		r.status = statusUndefined
		r.err = err
	case err == nil:
		r.status = statusOK
		r.err = nil
		r.collectedObject = obj
	case errors.As(err, &errRemoved):
		r.status = statusINHUMED
		r.err = errRemoved
	case errors.As(err, &errOutOfRange):
		r.status = statusOutOfRange
		r.err = errOutOfRange
	case errors.As(err, &errECInfo):
		r.status = statusEC
		r.err = err
	}
}
