package searchsvc

import (
	"context"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/client"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/container"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/netmap"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/engine"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/services/object/util"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/services/object_manager/placement"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/logger"
	cid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/id"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	"go.uber.org/zap"
)

// Service is an utility serving requests
// of Object.Search service.
type Service struct {
	*cfg
}

// Option is a Service's constructor option.
type Option func(*cfg)

type searchClient interface {
	// searchObjects searches objects on the specified node.
	// MUST NOT modify execCtx as it can be accessed concurrently.
	searchObjects(context.Context, *execCtx, client.NodeInfo) ([]oid.ID, error)
}

type ClientConstructor interface {
	Get(client.NodeInfo) (client.MultiAddressClient, error)
}

type cfg struct {
	log *logger.Logger

	localStorage interface {
		search(context.Context, *execCtx) ([]oid.ID, error)
	}

	clientConstructor interface {
		get(client.NodeInfo) (searchClient, error)
	}

	traverserGenerator interface {
		GenerateTraverser(cid.ID, *oid.ID, uint64) (*placement.Traverser, *container.Container, error)
	}

	currentEpochReceiver interface {
		Epoch() (uint64, error)
	}

	keyStore *util.KeyStorage

	containerSource container.Source
}

// New creates, initializes and returns utility serving
// Object.Get service requests.
func New(e *engine.StorageEngine,
	cc ClientConstructor,
	tg *util.TraverserGenerator,
	ns netmap.Source,
	ks *util.KeyStorage,
	cs container.Source,
	opts ...Option,
) *Service {
	c := &cfg{
		log: logger.NewLoggerWrapper(zap.L()),
		clientConstructor: &clientConstructorWrapper{
			constructor: cc,
		},
		localStorage: &storageEngineWrapper{
			storage: e,
		},
		traverserGenerator:   tg,
		currentEpochReceiver: ns,
		keyStore:             ks,
		containerSource:      cs,
	}

	for i := range opts {
		opts[i](c)
	}

	return &Service{
		cfg: c,
	}
}

// WithLogger returns option to specify Get service's logger.
func WithLogger(l *logger.Logger) Option {
	return func(c *cfg) {
		c.log = l.With(zap.String("component", "Object.Search service"))
	}
}
